structure LambdaCalculus =
struct
  val language_name = "lambda calculus"
  val evaluation_style = "lazy"

  datatype exp =
    Variable of string
  | Lambda of string * exp
  | Apply of exp * exp

  infix 6 ^^
  fun s1 ^^ s2 = s1 ^ " " ^ s2

  val rec pp_exp = (fn
      Variable x => x
    | Lambda (x, e) => "(fn" ^^ x ^^ "=>" ^^ pp_exp e ^ ")"
    | Apply (e1, e2) => "(" ^ pp_exp e1 ^^ pp_exp e2 ^ ")"
  )

  infix 4 free_in
  fun x free_in e =
    case e of
      Variable y => y = x
    | Lambda (y, e') => y <> x andalso x free_in e'
    | Apply (e1, e2) => x free_in e1 orelse x free_in e2

  fun avoid y bad =
    if bad y then avoid (y ^ "'") bad else y

  fun subst (x, e) = (fn
    Variable y => if y = x then e else Variable y
  | Lambda (y, e') =>
      if y = x then Lambda (y, e') else
      let val y' = avoid y (fn y' => (y' <> y andalso y' free_in e') orelse y' free_in e)
      in Lambda (y', (subst (x, e) o subst (y, Variable y')) e') end
  | Apply (e1, e2) => Apply (subst (x, e) e1, subst (x, e) e2)
  )

  (* small-step weak-head lazy dynamics *)
  val rec step = (fn
    Variable _ => NONE
  | Lambda (x, e) => NONE
  | Apply (e1, e2) => (
      case step e1 of SOME e1' => SOME (Apply (e1', e2)) | NONE => 
      case e1 of
        Lambda (x, body) => SOME (subst (x, e2) body)
      | _ => NONE
    )
  )
end
